#include <string>
#include <cstdio>
#include <cstdlib>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <GL/glut.h>
#include "GLUtility.h"

static GLUtility *my_glWidget = 0;
static std::string *my_libFilePath = 0;
static const int DefaultNumPts = 400;

static void
loadLibrary(const char *libraryPath)
{
    void *ret = ::dlopen(libraryPath, RTLD_NOW);
    if (ret == 0) {
        printf("Failed to load shared object: %s!\n", libraryPath);
        fflush(NULL);
    }
    else {
        printf("Loaded/Unloaded shared object successfully.\n");
        fflush(NULL);
        ::dlclose(ret);
    }
}


static void
initializeGL()
{
    glClearColor(0.61, 0.61, 1.0, 1.0);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_MULTISAMPLE);
    static GLfloat lightPosition[4] = { 0.5, 5.0, 7.0, 1.0 };
    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
}


static void
paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glTranslatef(0.0, 0.0, -10.0);
    glRotatef(0.0, 1.0, 0.0, 0.0);
    glRotatef(0.0, 0.0, 1.0, 0.0);
    glRotatef(0.0, 0.0, 0.0, 1.0);

    // Do the line drawing
    glEnableClientState(GL_VERTEX_ARRAY);
    vertIterator iter = my_glWidget->begin();
    for (; iter != my_glWidget->end(); ++iter) {
        vertArray tmp = *iter;
        glVertexPointer(3, GL_DOUBLE, 0, tmp.first);
        glDrawArrays(GL_LINE_STRIP, 0, tmp.second);
    }
    glDisableClientState(GL_VERTEX_ARRAY);
    
    glutSwapBuffers();
}


static void
resizeGL(int width, int height)
{
    int side = width < height ? width : height;
    glViewport((width - side) / 2, (height - side) / 2, side, side);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-0.5, +0.5, -0.5, +0.5, 4.0, 15.0);
    glMatrixMode(GL_MODELVIEW);
}


static void
handleKey(unsigned char key, int x, int y)
{
    switch (key) {
    case '1':
        if (0 != my_glWidget) {
            my_glWidget->createRandomLine();
            paintGL();
        }
        else {
            printf("Failed to draw lines!\n");
        }
        break;
    case '2':
        if (0 != my_libFilePath) {
            loadLibrary(my_libFilePath->c_str());
        }
        else {
            printf("Failed to load shared object!\n");
        }
        break;
    case '3':
        if (0 != my_glWidget) {
            my_glWidget->clearLines();
            paintGL();
        }
        else {
            printf("Failed to clear lines!\n");
        }
        break;
    case 27:
    case 'q':
        exit(0);
        break;
    }
}


int
main(int argc, char **argv)
{
    if (argc == 2) {
        printf("Library set as %s\n", argv[1]);
        my_libFilePath = new std::string(argv[1]);

        my_glWidget = new GLUtility(DefaultNumPts);

        glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA);
        glutInitWindowPosition(0, 0);
        glutInitWindowSize(400, 400);
        glutInit(&argc, argv);
        glutCreateWindow("glutSampleApp");

        initializeGL();
        glutReshapeFunc(resizeGL);
        glutDisplayFunc(paintGL);
        glutKeyboardFunc(handleKey);
        printf("\nGL Version: %s\n", glGetString(GL_VERSION));
        fflush(0);

        glutMainLoop();

        delete my_libFilePath;
        delete my_glWidget;
    }
    else {
        printf("\nA shared library MUST be specified.\n"
            "e.g. ./glutSampleApp /home/user/glutSampleApp/SimpleLibrary.so\n\n");
    }
    return 0;
}

